import polars as pl
import pytest

from wimsey import dataframe


def test_that_describe_returns_expected_dictionary_for_df() -> None:
    df = pl.DataFrame({"a": [1.2, 1.3, 1.4], "b": ["one", "two", None]})
    actual = dataframe._describe(df)
    assert 1.29 < actual["mean_a"] < 1.31
    assert actual["null_count_b"] == 1
    assert 0.332 < actual["null_percentage_b"] < 0.334
    assert actual["length"] == 3
    assert actual["columns"] == "a_^&^_b"


def test_that_describe_returns_expected_dictionary_for_lazy_frame() -> None:
    df = pl.LazyFrame({"a": [1.2, 1.3, 1.4], "b": ["one", "two", None]})
    actual = dataframe._describe(df)
    assert 1.29 < actual["mean_a"] < 1.31
    assert actual["null_count_b"] == 1
    assert 0.332 < actual["null_percentage_b"] < 0.334
    assert actual["length"] == 3
    assert actual["columns"] == "a_^&^_b"


def test_that_profile_by_sampling_returns_list_of_dicts_of_expected_length() -> None:
    df = pl.DataFrame({"a": [1.2, 1.3, 1.4], "b": ["one", "two", None]})
    actual = dataframe._profile_from_sampling(df, samples=10, n=1)
    assert len(actual) == 10
    assert actual[0]["mean_a"] in {1.2, 1.3, 1.4}
    assert actual[4]["columns"] == "a_^&^_b"


def test_that_profile_from_samples_returns_list_of_dicts_of_expected_length() -> None:
    dfs = [pl.DataFrame({"a": [1.2, 1.3, 1.4], "b": ["one", "two", None]}) for _ in range(20)]
    actual = dataframe._profile_from_samples(dfs)
    assert len(actual) == 20
    assert actual[10]["mean_a"] == 1.3
    assert actual[4]["columns"] == "a_^&^_b"


def test_that_describe_raises_error_for_empty_dataframe_with_no_columns() -> None:
    with pytest.raises(TypeError):
        dataframe._describe(pl.DataFrame())



import json
import textwrap

import yaml

from wimsey.tests import _possible_tests

arg_examples = {
    "column": "column_a",
    "other_column": "column_b",
    "be_less_than": 500,
    "be_less_than_or_equal_to": 300,
    "be_exactly": 300,
    "be_greater_than": 500,
    "be_greater_than_or_equal_to": 500,
    "have": ["column_a"],
    "not_have": ["column_c"],
    "be": ["column_a", "column_b"],
    "not_be": ["column_a", "column_b", "column_c"],
    "be_one_of": ["int64", "float64"],
    "not_be_one_of": ["a", "b"],
    "match_regex": "at$",
}


def generate_doc(name: str, doc_string: str, annotations: dict) -> str:
    examples = arg_examples
    if name == "type_should":
        examples["be"] = "int64"
        examples["not_be"] = "string"
    annotations.pop("return")
    try:
        dict_eg = {k: examples[k] for k in annotations}
    except KeyError:
        msg = (
            f"Argument for function {name} can't be generated because it is "
            "not in examples within generate_possible_tests arg_examples, consider"
            " adding.\n For reference here is the list of attempted examples: "
            f"{annotations}"
        )
        raise KeyError(msg) from KeyError
    yaml_eg = yaml.dump({"test": name} | dict_eg)
    json_eg = json.dumps({"test": name} | dict_eg, indent=2)
    python_eg = f"""
from wimsey import test
from wimsey.tests import {name}

keywords = {json.dumps(dict_eg, indent=2)}

result = test(df, contract=[{name}(**keywords)])
    """
    return f"""
## {name}

{doc_string}

=== "yaml"
    ```yaml
{textwrap.indent(yaml_eg, "    ")}
    ```
=== "json"
    ```json
{textwrap.indent(json_eg, "    ")}
    ```
=== "python"
    ```python
{textwrap.indent(python_eg, "    ")}
    ```

<hr>
    """


if __name__ == "__main__":
    NL = "\n"
    file_doc: str = "# Test Catalogue 🧪\n"
    file_doc += (
        "This documentation is intended as an exaustive list of possible "
        "tests within Wimsey. Note that examples given intentionally use "
        "all possible keywords for demonstrative purposes. This isn't "
        "required, and you can give as many or as few keywords as you like "
        "with the exception of where `column` is required."
    )
    for test_name, test_generator in _possible_tests.items():
        test_doc_string = test_generator.__doc__
        assert isinstance(test_doc_string, str), (
            f"{test_generator.__name__} has no docstring, this is required for doc generation"
        )
        doc = generate_doc(
            test_name,
            " ".join(test_generator.__doc__.replace("\n", "").split()),
            test_generator.__annotations__,
        )
        file_doc += doc
    with open("docs/possible-tests.md", "w") as file:
        file.write(file_doc)



"""Types (mostly data based) for use in Wimsey module."""

from dataclasses import dataclass
from typing import Any, Callable, TypeAlias

import narwhals.stable.v1 as nw


@dataclass
class _MagicExpr:
    """Type for use in place of nw.Expr.

    Special Wimsey object for special data such as column names that aren't
    possible via expressions directly.
    """

    expr_name: str


_schema = _MagicExpr("schema")


class DataValidationError(Exception):
    """Exception class for raising against invalid data."""


@dataclass
class Result:
    """Dataclass for storing test result information."""

    name: str
    success: bool
    unexpected: Any = None


@dataclass
class FinalResult:
    """Dataclass for storing test result information."""

    success: bool
    results: list[Result]


GeneratedTest: TypeAlias = tuple[nw.Expr | _MagicExpr, Callable[[Any], Result]]



from typing import Callable

import polars as pl

from wimsey import tests
from wimsey.execution import test as _test  # so pytest won't think this is test


def test_that_all_possible_tests_are_functions_that_return_partials() -> None:
    for test_name, actual_test in tests._possible_tests.items():
        assert isinstance(test_name, str)
        assert isinstance(
            actual_test(
                column="anything",
                other_column="anythin_else",
                be_less_than=3,
                be_one_of=[1, 2, 3],
            ),
            tuple,
        )


def test_all_possible_tests_exposed_as_variables_of_the_same_name_in_module() -> None:
    for test_name in tests._possible_tests:
        assert isinstance(getattr(tests, test_name), Callable)


def test_null_count_passes_and_fails_when_expected() -> None:
    df = pl.DataFrame({"a": [1, None, 2]})
    should_pass = _test(
        df,
        tests.null_count_should(
            column="a",
            be_exactly=1,
            be_less_than=2,
            be_less_than_or_equal_to=1,
            be_greater_than=0,
            be_greater_than_or_equal_to=1,
        ),
    )
    should_fail = _test(
        df,
        tests.null_count_should(
            column="a",
            be_exactly=2,
            be_less_than=1,
            be_less_than_or_equal_to=0,
            be_greater_than=1,
            be_greater_than_or_equal_to=0,
        ),
    )
    assert should_pass.success
    assert not should_fail.success


def test_null_percentage_passes_and_fails_when_expected() -> None:
    df = pl.DataFrame({"a": [1, None, 2, None]})
    should_pass = _test(
        df,
        tests.null_percentage_should(
            column="a",
            be_exactly=0.5,
            be_less_than=1,
            be_less_than_or_equal_to=0.5,
            be_greater_than=0,
            be_greater_than_or_equal_to=0.5,
        ),
    )
    should_fail = _test(
        df,
        tests.null_percentage_should(
            column="a",
            be_exactly=1,
            be_less_than=0,
            be_less_than_or_equal_to=0,
            be_greater_than=1,
            be_greater_than_or_equal_to=1,
        ),
    )
    assert should_pass.success
    assert not should_fail.success



import json
from collections.abc import Callable
from typing import NoReturn

import narwhals.stable.v1 as nw
import pytest
import yaml

from wimsey import config
from wimsey.types import _MagicExpr


@pytest.fixture
def test_suite():
    return [
        {"test": "mean_should", "column": "a", "be_exactly": 34},
        {"test": "type_should", "column": "y", "be_one_of": ["int64", "float64"]},
    ]


def throw_import_error(*args, **kwargs) -> NoReturn:
    raise ImportError


def test_collect_tests_returns_list_of_tuples_with_expressions_and_callables(
    test_suite,
) -> None:
    actual = config._collect_tests(test_suite)
    for expr, callable in actual:
        assert isinstance(expr, (nw.Expr, _MagicExpr))
        assert isinstance(callable, Callable)


def test_collect_tests_returns_friendly_error_when_required_value_not_give(test_suite) -> None:
    test_suite[0].pop("column")
    with pytest.raises(TypeError, match="column"):
        config._collect_tests(test_suite)


def test_collect_tests_returns_friendly_error_when_no_test_is_given(test_suite) -> None:
    test_suite[1].pop("test")
    with pytest.raises(ValueError, match="test"):
        config._collect_tests(test_suite)


def test_collect_tests_returns_input_when_input_is_already_test_functions(test_suite) -> None:
    initial = config._collect_tests(test_suite)
    actual = config._collect_tests(initial)
    assert actual == initial


def test_read_config_parses_yaml(monkeypatch, test_suite) -> None:
    class DummyOpenFile:
        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs): ...

        def read(self, *args, **kwargs):
            return yaml.dump(test_suite)

    def open_file_patch(*args, **kwargs):
        return DummyOpenFile()

    monkeypatch.setattr(config.fsspec, "open", open_file_patch)
    actual = config._read_config("file.yaml")
    for expr, callable in actual:
        assert isinstance(expr, (nw.Expr, _MagicExpr))
        assert isinstance(callable, Callable)


def test_read_config_parses_yaml_with_test_section(monkeypatch, test_suite) -> None:
    class DummyOpenFile:
        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs): ...

        def read(self, *args, **kwargs):
            return yaml.dump({"cool": ["some", "cool", "stuff"], "tests": test_suite})

    def open_file_patch(*args, **kwargs):
        return DummyOpenFile()

    monkeypatch.setattr(config.fsspec, "open", open_file_patch)
    actual = config._read_config("file.yaml")
    for expr, callable in actual:
        assert isinstance(expr, (nw.Expr, _MagicExpr))
        assert isinstance(callable, Callable)


def test_read_config_parses_yaml_with_only_one_test(monkeypatch, test_suite) -> None:
    class DummyOpenFile:
        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs): ...

        def read(self, *args, **kwargs):
            return yaml.dump(test_suite[0])

    def open_file_patch(*args, **kwargs):
        return DummyOpenFile()

    monkeypatch.setattr(config.fsspec, "open", open_file_patch)
    actual = config._read_config("file.yaml")
    for expr, callable in actual:
        assert isinstance(expr, (nw.Expr, _MagicExpr))
        assert isinstance(callable, Callable)


def test_read_config_parses_json(monkeypatch, test_suite) -> None:
    class DummyOpenFile:
        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs): ...

        def read(self, *args, **kwargs):
            return json.dumps(test_suite)

    def open_file_patch(*args, **kwargs):
        return DummyOpenFile()

    monkeypatch.setattr(config.fsspec, "open", open_file_patch)
    actual = config._read_config("file.json")
    for expr, callable in actual:
        assert isinstance(expr, (nw.Expr, _MagicExpr))
        assert isinstance(callable, Callable)


def test_friendly_message_is_raised_when_yaml_is_unimportable(test_suite, monkeypatch) -> None:
    class DummyOpenFile:
        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs): ...

        def read(self, *args, **kwargs):
            return json.dumps(test_suite)

    def open_file_patch(*args, **kwargs):
        return DummyOpenFile()

    monkeypatch.setattr(config.fsspec, "open", open_file_patch)
    monkeypatch.setattr(config, "_collect_tests", throw_import_error)
    with pytest.raises(ImportError, match="pip install pyyaml"):
        config._read_config("file.yaml")


def test_friendly_message_is_raised_when_yaml_does_not_return_contents(
    test_suite,
    monkeypatch,
) -> None:
    class DummyOpenFile:
        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs): ...

        def read(self, *args, **kwargs) -> str:
            return "dsafasdfasdf"

    def open_file_patch(*args, **kwargs):
        return DummyOpenFile()

    monkeypatch.setattr(config.fsspec, "open", open_file_patch)
    with pytest.raises(ValueError, match="json/yaml"):
        config._read_config("file.yaml")



"""Functions for profiling and building tests."""

from __future__ import annotations

import json
from enum import Enum, auto
from statistics import stdev
from typing import TYPE_CHECKING, Any

import fsspec

from wimsey.dataframe import _profile_from_samples, _profile_from_sampling
from wimsey.execution import test, validate
from wimsey.types import FinalResult, Result

if TYPE_CHECKING:
    from collections.abc import Iterable

    from narwhals.stable.v1.typing import FrameT


class _StarterTestStatus(Enum):
    """Internal class to mark out column consistency for use."""

    UNSET = auto()
    SET = auto()
    CANCELLED = auto()


def starter_tests_from_samples(
    samples: Iterable[FrameT],
    margin: float = 1,
) -> list[dict]:
    """From a list of supported dataframes, produce a list of passing tests.

    Margin is worth explaining here, as it's the amount of *extra margin* tests
    are given, based on standard deviation. If three dataframes are given, with
    "column_a" values that have maximums of 1, 2 and 3, rather than creating a
    test that the maximum should be 3, Wimsey will give a degree of margin.

    By default this will be 1 standard deviation (of the maximum values). So for
    the above example, Wimsey will test for a maximum for 4. This can be tuned with
    the 'margin' keyword.
    """
    df_samples: list[dict[str, Any]] = _profile_from_samples(samples)
    return _starter_tests_from_sample_describes(df_samples, margin)


def save_starter_tests_from_samples(
    path: str,
    samples: Iterable[FrameT],
    margin: float = 1,
    storage_options: dict | None = None,
) -> None:
    """Build and save starter tests from iterable of dataframe.

    See starter_tests_from_samples for more information, will additionally
    save tests as yaml or json dependend on path extension.
    """
    storage_options = storage_options or {}
    starter_tests = starter_tests_from_samples(samples, margin)
    _save_starter_tests(path, starter_tests, storage_options=storage_options)


def save_starter_tests_from_sampling(
    path: str,
    df: FrameT,
    samples: int = 100,
    n: int | None = None,
    fraction: int | None = None,
    margin: float = 1,
    storage_options: dict | None = None,
) -> None:
    """Create and save starter tests from dataframe.

    See starter_tests_from_sampling for more information, will additionally
    save tests as yaml or json dependend on path extension.
    """
    storage_options = storage_options or {}
    starter_tests = starter_tests_from_sampling(
        df=df,
        samples=samples,
        n=n,
        fraction=fraction,
        margin=margin,
    )
    _save_starter_tests(path, starter_tests, storage_options=storage_options)


def _save_starter_tests(
    path: str,
    tests: list[dict],
    storage_options: dict | None = None,
) -> None:
    if path.endswith((".yaml", ".yml")):
        try:
            import yaml
        except ImportError as exception:  # pragma: no cover
            msg = (
                "It looks like you're trying to import a yaml configured "
                "test suite. This is supported but requires an additional "
                "install of pyyaml (`pip install pyyaml`)"
            )
            raise ImportError(msg) from exception
        contents = yaml.dump(tests)
    else:
        contents = json.dumps(tests)
    with fsspec.open(path, mode="wt", **storage_options) as file:
        file.write(contents)


def starter_tests_from_sampling(
    df: FrameT,
    samples: int = 100,
    n: int | None = None,
    fraction: int | None = None,
    margin: float = 1,
) -> list[dict]:
    """Build out starter tests from a dataframe.

    From a supported dataframe, produce a list of passing tests by sampling the
    dataframe. Note n *or* fraction should be given, but not both. Keyword `samples`
    relates to the *number of samples* to take, whereas `n` relates to the *size of
    a given sample*.

    Margin is worth explaining here, as it's the amount of *extra margin* tests
    are given, based on standard deviation. If three dataframes are given, with
    "column_a" values that have maximums of 1, 2 and 3, rather than creating a
    test that the maximum should be 3, Wimsey will give a degree of margin.

    By default this will be 1 standard deviation (of the maximum values). So for
    the above example, Wimsey will test for a maximum for 4. This can be tuned with
    the 'margin' keyword.
    """
    df_samples = _profile_from_sampling(df, samples, n, fraction)
    return _starter_tests_from_sample_describes(df_samples, margin)


def _starter_tests_from_sample_describes(
    samples: list[dict],
    margin: float = 1,
) -> list[dict]:
    """Build out starter tests from `describe` dictionary returs."""
    column_test: dict = {"test": "columns_should", "status": _StarterTestStatus.UNSET}
    for sample in samples:
        column_test = _update_column_starter_test(column_test, sample)
    if column_test["status"] is _StarterTestStatus.CANCELLED:  # pragma: no cover
        msg = (
            "There aren't any consistently held columns in the samples "
            "so Wimsey is unable to build a start test from them."
        )
        raise ValueError(msg)
    column_test.pop("status")
    if column_test.get("be") is None:  # pragma: no cover
        return [column_test]
    column_tests = _type_starter_tests(samples, column_test["be"])
    for stat in ["mean", "std", "max", "min", "null_percentage"]:
        column_tests += _stat_starter_tests(
            stat,
            samples,
            column_test["be"],
            margin=margin,
        )
    return [*column_tests, column_test]


def _update_column_starter_test(starter: dict, sample: dict) -> dict:
    """Update column checks for a single sample."""
    sample_columns = set(sample["columns"].split("_^&^_"))
    if starter["status"] is _StarterTestStatus.UNSET:
        starter["be"] = list(sample_columns)
        starter["status"] = _StarterTestStatus.SET
    elif starter["status"] is _StarterTestStatus.CANCELLED:  # pragma: no cover
        pass
    elif starter.get("be") is not None and set(starter["be"]) != set(
        sample_columns,
    ):  # pragma: no cover
        old_be = starter.pop("be")
        new_have = set(old_be) & set(sample_columns)
        if len(new_have) > 0:
            starter["have"] = list(new_have)
        else:
            starter["status"] = _StarterTestStatus.CANCELLED
    return starter


def _stat_starter_tests(
    stat: str,
    samples: list[dict],
    columns: list[str],
    margin: float,
) -> list[dict]:
    """Build out statistical starter tests."""
    tests: list[dict] = []
    for column in columns:
        values = [i[f"{stat}_{column}"] for i in samples if i[f"{stat}_{column}"] is not None]
        if len(values) == 0:
            continue
        absolute_margin = stdev(values) * margin
        if absolute_margin == 0:
            test = {
                "column": column,
                "test": f"{stat}_should",
                "be_exactly": max(values),
            }
        else:
            test = {
                "column": column,
                "test": f"{stat}_should",
                "be_less_than_or_equal_to": max(values) + absolute_margin,
                "be_greater_than_or_equal_to": min(values) - absolute_margin,
            }
        tests.append(test)
    return tests


def _type_starter_tests(
    samples: list[dict],
    columns: list[str],
) -> list[dict]:
    """Build out type based starter tests."""
    tests: list[dict] = []
    for column in columns:
        types: set[str] = {i[f"type_{column}"] for i in samples}
        test: dict[str, Any] = {"column": column, "test": "type_should"}
        if len(types) == 1:
            test |= {"be": next(iter(types))}
        else:  # pragma: no cover
            test |= {"be_one_of": list(types)}
        tests.append(test)
    return tests


def validate_or_build(
    df: FrameT,
    contract: str,
    samples: int = 100,
    n: int | None = None,
    fraction: int | None = None,
    margin: float = 1,
    storage_options: dict | None = None,
) -> FrameT:
    """Validate dataframe if test file exists, otherwise make one.

    Test file will be generated from multple samples of dataset.

    Will fall back to starter_tests_from_sampling (a list samples is not possible with
    only one dataframe), see *starter_tests_from_sampling* and *save_starter_tests_from_sampling*
    for more details on use of keyword arguments aside from df, contract and storage_options.
    """
    try:
        return validate(df, contract=contract, storage_options=storage_options)
    except FileNotFoundError:
        save_starter_tests_from_sampling(
            path=contract,
            df=df,
            samples=samples,
            n=n,
            fraction=fraction,
            margin=margin,
        )
        return df


def test_or_build(
    df: FrameT,
    contract: str,
    samples: int = 100,
    n: int | None = None,
    fraction: int | None = None,
    margin: float = 1,
    storage_options: dict | None = None,
) -> FinalResult:
    """Validate dataframe if test file exists, otherwise make one.

    Test file will be generated from multple samples of dataset.

    Will fall back to starter_tests_from_sampling (a list samples is not possible with
    only one dataframe), see *starter_tests_from_sampling* and *save_starter_tests_from_sampling*
    for more details on use of keyword arguments aside from df, contract and storage_options.
    """
    try:
        return test(df, contract=contract, storage_options=storage_options)
    except FileNotFoundError:
        save_starter_tests_from_sampling(
            path=contract,
            df=df,
            samples=samples,
            n=n,
            fraction=fraction,
            margin=margin,
        )
        return FinalResult(True, [Result("tests-generated", True)])



"""Wimsey test definitions.

Definitions for test functions that can be ran against a dataframe.
"""

from __future__ import annotations

from typing import Any, Callable

import narwhals.stable.v1 as nw

from wimsey.types import GeneratedTest, Result, _schema


def _range_check(
    aggregation: Callable[[str], nw.Expr],
    metric_name: str,
) -> Callable:
    """Generate tests of the form "x should be within range".

    Tests are also factories in themselves, they'll generate functions to take
    only a "describe" object.
    """

    def check_aggregation_is_in_range(
        column: str,
        be_exactly: float | None = None,
        be_less_than: float | None = None,
        be_less_than_or_equal_to: float | None = None,
        be_greater_than: float | None = None,
        be_greater_than_or_equal_to: float | None = None,
        **kwargs: Any,
    ) -> GeneratedTest:
        """Test that column metric is within designated range."""
        del kwargs

        def _check(value: float) -> Result:
            checks = []
            if be_exactly is not None:
                checks.append(value == be_exactly)
            if be_less_than is not None:
                checks.append(value < be_less_than)
            if be_greater_than is not None:
                checks.append(value > be_greater_than)
            if be_less_than_or_equal_to is not None:
                checks.append(value <= be_less_than_or_equal_to)
            if be_greater_than_or_equal_to is not None:
                checks.append(value >= be_greater_than_or_equal_to)
            return Result(
                name=f"{metric_name}-of-{column}",
                success=all(checks),
                unexpected=value if not all(checks) else None,
            )

        return aggregation(column), _check

    return check_aggregation_is_in_range


def row_count_should(
    be_less_than: float | None = None,
    be_less_than_or_equal_to: float | None = None,
    be_greater_than: float | None = None,
    be_greater_than_or_equal_to: float | None = None,
    be_exactly: float | None = None,
    **kwargs: Any,
) -> GeneratedTest:
    """Test that dataframe row count is within designated range."""
    del kwargs

    def _check(value: float) -> Result:
        """Test that dataframe row count is within designated range."""
        checks: list[bool] = []
        if be_exactly is not None:
            checks.append(value == be_exactly)
        if be_less_than is not None:
            checks.append(value < be_less_than)
        if be_greater_than is not None:
            checks.append(value > be_greater_than)
        if be_less_than_or_equal_to is not None:
            checks.append(value <= be_less_than_or_equal_to)
        if be_greater_than_or_equal_to is not None:
            checks.append(value >= be_greater_than_or_equal_to)
        return Result(
            name="row-count",
            success=all(checks),
            unexpected=value if not all(checks) else None,
        )

    return nw.len(), _check


def average_difference_from_other_column_should(
    column: str,
    other_column: str,
    be_exactly: float | None = None,
    be_less_than: float | None = None,
    be_less_than_or_equal_to: float | None = None,
    be_greater_than: float | None = None,
    be_greater_than_or_equal_to: float | None = None,
    **kwargs: Any,
) -> GeneratedTest:
    """Compare difference between two columns.

    Test that the average difference between column and other column are
    within designated bounds.
    """
    del kwargs

    def _check(difference: float) -> Result:
        checks: list[bool] = []
        if be_exactly is not None:
            checks.append(difference == be_exactly)
        if be_less_than is not None:
            checks.append(difference < be_less_than)
        if be_less_than_or_equal_to is not None:
            checks.append(difference <= be_less_than_or_equal_to)
        if be_greater_than is not None:
            checks.append(difference > be_greater_than)
        if be_greater_than_or_equal_to is not None:
            checks.append(difference >= be_greater_than_or_equal_to)
        return Result(
            name=f"average-difference-from-{column}-to-{other_column}",
            success=all(checks),
            unexpected=difference if not all(checks) else None,
        )

    return (nw.col(column) - nw.col(other_column)).mean(), _check


def average_ratio_to_other_column_should(
    column: str,
    other_column: str,
    be_exactly: float | None = None,
    be_less_than: float | None = None,
    be_less_than_or_equal_to: float | None = None,
    be_greater_than: float | None = None,
    be_greater_than_or_equal_to: float | None = None,
    **kwargs: Any,
) -> GeneratedTest:
    """Compare ratio between two columns.

    Test that the average ratio between column and other column are
    within designated bounds (for instance, a value of 1 has a ratio
    of 0.1 to a value of 10).
    """
    del kwargs

    def _check(ratio: float) -> Result:
        checks: list[bool] = []
        if be_exactly is not None:
            checks.append(ratio == be_exactly)
        if be_less_than is not None:
            checks.append(ratio < be_less_than)
        if be_less_than_or_equal_to is not None:
            checks.append(ratio <= be_less_than_or_equal_to)
        if be_greater_than is not None:
            checks.append(ratio > be_greater_than)
        if be_greater_than_or_equal_to is not None:
            checks.append(ratio >= be_greater_than_or_equal_to)
        return Result(
            name=f"average-ratio-between-{column}-and-{other_column}",
            success=all(checks),
            unexpected=ratio if not all(checks) else None,
        )

    return (nw.col(column) / nw.col(other_column)).mean(), _check


def max_string_length_should(
    column: str,
    be_exactly: float | None = None,
    be_less_than: float | None = None,
    be_less_than_or_equal_to: float | None = None,
    be_greater_than: float | None = None,
    be_greater_than_or_equal_to: float | None = None,
    **kwargs: Any,
) -> GeneratedTest:
    """Test that the maximum string length is within expected bounds."""
    del kwargs

    def _check(max_length: float) -> Result:
        checks: list[bool] = []
        if be_exactly is not None:
            checks.append(max_length == be_exactly)
        if be_less_than is not None:
            checks.append(max_length < be_less_than)
        if be_less_than_or_equal_to is not None:
            checks.append(max_length <= be_less_than_or_equal_to)
        if be_greater_than is not None:
            checks.append(max_length > be_greater_than)
        if be_greater_than_or_equal_to is not None:
            checks.append(max_length >= be_greater_than_or_equal_to)
        return Result(
            name=f"average-max-string-length-of-{column}",
            success=all(checks),
            unexpected=max_length if not all(checks) else None,
        )

    return nw.col(column).str.len_chars().max(), _check


def all_values_should(
    column: str,
    be_one_of: list[str] | None = None,
    not_be_one_of: list[str] | None = None,
    match_regex: str | None = None,
    **kwargs: Any,
) -> GeneratedTest:
    """Test all unique values within a column are within expected group."""
    del kwargs

    def _check(
        success: bool,
    ) -> Result:
        return Result(
            name=f"all-values-of-{column}",
            success=success,
            unexpected=None if success else "Values did not meet given conditions",
        )

    expressions: list[nw.Expr] = []
    if be_one_of:
        expressions.append(nw.col(column).is_in(be_one_of).min())
    if not_be_one_of:
        expressions.append(~(nw.col(column).is_in(not_be_one_of)).max())
    if match_regex:
        expressions.append(nw.col(column).str.contains(match_regex))
    return nw.all_horizontal(*expressions), _check


def type_should(
    column: str,
    be: str | None = None,
    not_be: str | None = None,
    be_one_of: list[str] | None = None,
    **kwargs: Any,
) -> GeneratedTest:
    """Check for type of column.

    Check that type of column meets expected criteria. Note that because Wimsey
    is a dataframe agnostic tool, this should be of *Narwhals* expected types,
    such as Float64, Int64, String, etc.

    See Narwhals' documentation for more details:
    https://narwhals-dev.github.io/narwhals/api-reference/dtypes/
    """
    del kwargs

    def _check(schema_dict: dict) -> Result:
        col_type = schema_dict[column]
        checks: list[bool] = []
        if be is not None:
            checks.append(be.lower() == col_type.lower())
        if not_be is not None:
            checks.append(not_be.lower() != col_type.lower())
        if be_one_of is not None:
            checks.append(col_type.lower() in [i.lower() for i in be_one_of])
        return Result(
            name=f"type-of-{column}",
            success=all(checks),
            unexpected=col_type if not all(checks) else None,
        )

    return _schema, _check


def columns_should(
    have: list[str] | str | None = None,
    not_have: list[str] | str | None = None,
    be: list[str] | str | None = None,
    **kwargs: Any,
) -> GeneratedTest:
    """Check that expected columns are present / non-present within dataframe."""
    del kwargs

    def _check(schema_dict: dict) -> Result:
        _have = list(have) if isinstance(have, str) else have
        _not_have = list(not_have) if isinstance(not_have, str) else not_have
        _be = list(be) if isinstance(be, str) else be
        checks: list[bool] = []
        present_columns = list(schema_dict)
        if _have is not None:
            checks.extend(col in present_columns for col in _have)
        if _not_have is not None:
            checks.extend(col not in present_columns for col in _not_have)
        if _be is not None:
            checks.append(set(present_columns) == set(_be))
            checks.append(len(present_columns) == len(_be))
        return Result(
            name="columns",
            success=all(checks),
            unexpected=present_columns if not all(checks) else None,
        )

    return _schema, _check


def null_count_should(
    column: str,
    be_exactly: float | None = None,
    be_less_than: float | None = None,
    be_less_than_or_equal_to: float | None = None,
    be_greater_than: float | None = None,
    be_greater_than_or_equal_to: float | None = None,
    **kwargs: Any,
) -> GeneratedTest:
    """Check that null count of column meets expected criteria."""
    del kwargs

    def _check(null_count: float) -> Result:
        checks = []
        if be_exactly is not None:
            checks.append(null_count == be_exactly)
        if be_less_than is not None:
            checks.append(null_count < be_less_than)
        if be_greater_than is not None:
            checks.append(null_count > be_greater_than)
        if be_less_than_or_equal_to is not None:
            checks.append(null_count <= be_less_than_or_equal_to)
        if be_greater_than_or_equal_to is not None:
            checks.append(null_count >= be_greater_than_or_equal_to)
        return Result(
            name=f"null-count-of-{column}",
            success=all(checks),
            unexpected=null_count if not all(checks) else None,
        )

    return nw.col(column).null_count(), _check


def null_percentage_should(
    column: str,
    be_exactly: float | None = None,
    be_less_than: float | None = None,
    be_less_than_or_equal_to: float | None = None,
    be_greater_than: float | None = None,
    be_greater_than_or_equal_to: float | None = None,
    **kwargs: Any,
) -> GeneratedTest:
    """Check that null percentage of column meets expected criteria."""
    del kwargs

    def _check(null_percentage: float) -> Result:
        checks = []
        if be_exactly is not None:
            checks.append(null_percentage == be_exactly)
        if be_less_than is not None:
            checks.append(null_percentage < be_less_than)
        if be_greater_than is not None:
            checks.append(null_percentage > be_greater_than)
        if be_less_than_or_equal_to is not None:
            checks.append(null_percentage <= be_less_than_or_equal_to)
        if be_greater_than_or_equal_to is not None:
            checks.append(null_percentage >= be_greater_than_or_equal_to)
        return Result(
            name=f"null-percentage-of-{column}",
            success=all(checks),
            unexpected=null_percentage if not all(checks) else None,
        )

    return (nw.col(column).null_count() / nw.col(column).len()), _check


_possible_tests: dict[str, Callable] = {
    "mean_should": (mean_should := _range_check(nw.mean, "mean")),
    "min_should": (min_should := _range_check(nw.min, "min")),
    "max_should": (max_should := _range_check(nw.max, "max")),
    "std_should": (std_should := _range_check(lambda col: nw.col(col).std(), "stdev")),
    "count_should": (count_should := _range_check(lambda col: nw.col(col).count(), "count")),
    "row_count_should": row_count_should,
    "average_difference_from_other_column_should": average_difference_from_other_column_should,
    "average_ratio_to_other_column_should": average_ratio_to_other_column_should,
    "max_string_length_should": max_string_length_should,
    "all_values_should": all_values_should,
    "type_should": type_should,
    "columns_should": columns_should,
    "null_count_should": null_count_should,
    "null_percentage_should": null_percentage_should,
}



import polars as pl
import pytest

from wimsey import execution, tests


def test_run_all_tests_produces_expected_result_object() -> None:
    tests_to_carry_out = [
        tests.max_should(column="a", be_less_than=10),
        tests.std_should(column="a", be_greater_than=0),
        tests.type_should(column="b", be_one_of=["string", "int64"]),
    ]
    df = pl.DataFrame({"a": [1, 2, 3], "b": ["hat", "bat", "cat"]})
    actual = execution._run_all_tests(df, tests_to_carry_out)
    assert actual.success is True
    for result in actual.results:
        assert result.success is True


def test_validate_carries_out_tests_then_returns_object_if_passing() -> None:
    tests_to_carry_out = [
        tests.max_should(column="a", be_less_than=10),
        tests.std_should(column="a", be_greater_than=0),
        tests.type_should(column="b", be_one_of=["string", "int64"]),
    ]
    df = pl.DataFrame({"a": [1, 2, 3], "b": ["hat", "bat", "cat"]})
    actual = execution.validate(df, tests_to_carry_out)
    assert isinstance(actual, pl.DataFrame)


def test_validate_raises_error_if_tests_fail() -> None:
    tests_to_carry_out = [
        tests.max_should(column="a", be_exactly=0),
        tests.std_should(column="a", be_greater_than=10),
        tests.type_should(column="b", be_one_of=["string", "int64"]),
    ]
    df = pl.DataFrame({"a": [1, 2, 3], "b": ["hat", "bat", "cat"]})
    with pytest.raises(execution.DataValidationError):
        execution.validate(df, tests_to_carry_out)


def test_row_count_expectations_pass_when_expected() -> None:
    tests_to_carry_out = [
        tests.row_count_should(
            be_less_than=3.1,
            be_less_than_or_equal_to=3,
            be_greater_than=-2.343,
            be_greater_than_or_equal_to=0.3,
            be_exactly=3,
        ),
    ]
    df = pl.DataFrame({"a": [1, 2, 3], "b": ["hat", "bat", "cat"]})
    actual = execution._run_all_tests(df, tests_to_carry_out)
    assert actual.success is True
    for result in actual.results:
        assert result.success is True


def test_columns_should_have_expectations_fail_when_expected() -> None:
    tests_to_carry_out = [
        tests.columns_should(
            have="c",
            not_have="a",
            be=["b", "c"],
        ),
    ]
    df = pl.DataFrame({"a": [1, 2, 3], "b": ["hat", "bat", "cat"]})
    actual = execution._run_all_tests(df, tests_to_carry_out)
    assert actual.success is False


def test_column_type_tests_pass_when_expected() -> None:
    tests_to_carry_out = [
        tests.type_should(column="a", be="int64"),
        tests.type_should(column="a", be_one_of=["int64", "float64"]),
        tests.type_should(column="a", not_be="float64"),
        tests.type_should(column="b", be="string"),
    ]
    df = pl.DataFrame({"a": [1, 2, 3], "b": ["hat", "bat", "cat"]})
    actual = execution._run_all_tests(df, tests_to_carry_out)
    assert actual.success is True
    for result in actual.results:
        assert result.success is True


def test_average_column_difference_tests_pass_when_expected() -> None:
    tests_to_carry_out = [
        tests.average_difference_from_other_column_should(
            column="a",
            other_column="b",
            be_exactly=0,
            be_less_than=2,
            be_greater_than=-1,
            be_less_than_or_equal_to=0,
            be_greater_than_or_equal_to=0,
        ),
    ]
    df = pl.DataFrame({"a": [1, 2, 3], "b": [1, 2, 3]})
    actual = execution._run_all_tests(df, tests_to_carry_out)
    assert actual.success is True
    for result in actual.results:
        assert result.success is True


def test_average_column_ratio_tests_pass_when_expected() -> None:
    tests_to_carry_out = [
        tests.average_ratio_to_other_column_should(
            column="b",
            other_column="a",
            be_exactly=2,
            be_less_than=3,
            be_greater_than=1,
            be_less_than_or_equal_to=2,
            be_greater_than_or_equal_to=2,
        ),
    ]
    df = pl.DataFrame({"a": [1, 2, 3], "b": [2, 4, 6]})
    actual = execution._run_all_tests(df, tests_to_carry_out)
    assert actual.success is True
    for result in actual.results:
        assert result.success is True


def test_max_string_length_tests_fail_when_expected() -> None:
    tests_to_carry_out = [
        tests.max_string_length_should(
            column="a",
            be_less_than=4,
            be_less_than_or_equal_to=4,
            be_greater_than=2,
            be_greater_than_or_equal_to=2,
            be_exactly=3.0,
        ),
    ]
    df = pl.DataFrame({"a": ["cat", "bat", "ssssssssssssssnake"]})
    actual = execution._run_all_tests(df, tests_to_carry_out)
    assert actual.success is False


def test_all_values_tests_succeed_when_expected() -> None:
    tests_to_carry_out = [
        tests.all_values_should(
            column="a",
            be_one_of=["cat", "bat", "mat"],
            not_be_one_of=["hat"],
            match_regex="at$",
        ),
    ]
    df = pl.DataFrame({"a": ["cat", "bat"]})
    actual = execution._run_all_tests(df, tests_to_carry_out)
    assert actual.success is True
    for result in actual.results:
        assert result.success is True



from typing import Callable

import polars as pl
import pytest

from wimsey import execution, profiling
from wimsey.types import FinalResult


def raise_exception_patch(exception_type: type[Exception]) -> Callable:
    """Create a patch that will throw an error."""

    def raise_exception(*args, **kwargs) -> None:
        raise exception_type

    return raise_exception


def do_nothing(*args, **kwargs) -> None: ...


def test_starter_tests_from_sampling_returns_passing_test() -> None:
    df = pl.DataFrame(
        {
            "a": [1, 2, 3, 4, 5],
            "b": ["hat", "bat", "cat", "mat", "sat"],
            "c": [0.2, 0.4, 0.2, 0.56, 0.1],
        },
    )
    starter_test = profiling.starter_tests_from_sampling(df, samples=100, n=5)
    result = execution.test(df, starter_test)
    assert result.success


def test_starter_tests_from_samples_returns_passing_test() -> None:
    df = pl.DataFrame(
        {
            "a": [1, 2, 3, 4, 5],
            "b": ["hat", "bat", "cat", "mat", "sat"],
            "c": [0.2, 0.4, 0.2, 0.56, 0.1],
        },
    )
    starter_test = profiling.starter_tests_from_samples(
        [df.sample(fraction=0.5) for _ in range(100)],
    )
    result = execution.test(df, starter_test)
    assert result.success


def test_margin_works_as_anticipated() -> None:
    df = pl.DataFrame(
        {
            "a": [1, 2, 3, 4, 5],
            "b": ["hat", "bat", "cat", "mat", "sat"],
            "c": [0.2, 0.4, 0.2, 0.56, 0.1],
        },
    )
    starter_test = profiling.starter_tests_from_sampling(df, n=5, margin=50)
    result = execution.test(df, starter_test)
    assert result.success
    impossible_test = profiling.starter_tests_from_sampling(df, n=5, margin=-500)
    result = execution.test(df, impossible_test)
    assert not result.success


def test_save_tests_from_sampling_creates_expected_and_runnable_file(tmp_path) -> None:
    df = pl.DataFrame(
        {
            "a": [1, 2, 3, 4, 5],
            "b": ["hat", "bat", "cat", "mat", "sat"],
            "c": [0.2, 0.4, 0.2, 0.56, 0.1],
        },
    )
    profiling.save_starter_tests_from_sampling(
        str(tmp_path / "cool.yaml"),
        df,
        n=5,
        margin=1,
    )
    result = execution.test(df, str(tmp_path / "cool.yaml"))
    assert result.success


def test_save_tests_from_samples_creates_expected_and_runnable_file(tmp_path) -> None:
    df = pl.DataFrame(
        {
            "a": [1, 2, 3, 4, 5],
            "b": ["hat", "bat", "cat", "mat", "sat"],
            "c": [0.2, 0.4, 0.2, 0.56, 0.1],
        },
    )
    profiling.save_starter_tests_from_samples(
        str(tmp_path / "cool.json"),
        [df.sample(fraction=0.5) for _ in range(10)],
    )
    result = execution.test(df, str(tmp_path / "cool.json"))
    assert result.success


def test_validate_or_build_falls_back_to_save_starter_tests_from_sampling_if_validate_crashes(
    monkeypatch,
) -> None:
    monkeypatch.setattr(profiling, "validate", raise_exception_patch(FileNotFoundError))
    monkeypatch.setattr(
        profiling,
        "save_starter_tests_from_sampling",
        do_nothing,
    )
    actual = profiling.validate_or_build(3, "cool.json")
    assert actual == 3


def test_test_or_build_falls_back_to_save_starter_tests_from_sampling_if_validate_crashes(
    monkeypatch,
) -> None:
    monkeypatch.setattr(profiling, "test", raise_exception_patch(FileNotFoundError))
    monkeypatch.setattr(
        profiling,
        "save_starter_tests_from_sampling",
        do_nothing,
    )
    actual = profiling.test_or_build(3, "cool.json")
    assert isinstance(actual, FinalResult)
    assert actual.success



"""Functions for dataframe interactions, like sampling and executing expressions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import narwhals.stable.v1 as nw

if TYPE_CHECKING:
    from collections.abc import Iterable

    from narwhals.stable.v1 import NwDataFrame
    from narwhals.stable.v1.typing import IntoFrame


def _narwhals_to_dict(df: NwDataFrame) -> dict:
    """Convert Narwhals datframe to dict.

    Private function to convert Narwhals' FrameT to dict,
    note that this will *only* present the first row because
    it is built for specific evaluation of aggergate values.

    As such it's not suitable for general use.

    Also note, this will *force computation* for lazy frames,
    this is needed in order for Wimsey to carry out certain
    tests.
    """
    try:
        return {k: v[0] for k, v in df.to_dict(as_series=False).items()}  # type: ignore[union-attr]
    except AttributeError:
        return {
            k: v[0]
            for k, v in df.collect().to_dict(as_series=False).items()  # type: ignore[union-attr]
        }


def _validate_df_has_columns(df: NwDataFrame) -> NwDataFrame:
    if not df.columns:
        msg = "Wimsey cannot build a profile of this dataframe because it has no columns"
        raise TypeError(msg)
    return df


def _evaluate(
    df: IntoFrame,
    metrics: list[nw.Expr],
) -> dict[str, Any]:
    """Execute a list of scalar expressions and return dictionary of results."""
    nw_df: NwDataFrame = nw.from_native(df)
    metrics += [nw.lit(str(nw_df.schema[c])).alias(f"type_{c}") for c in nw_df.columns]

    evaluation_df = nw_df.pipe(_validate_df_has_columns).select(*metrics)
    evaluation_dict = _narwhals_to_dict(evaluation_df)
    type_evals: list = [i for i in evaluation_dict if i.startswith("type")]
    schema_dict: dict = {i.replace("type_", ""): evaluation_dict.pop(i) for i in type_evals}
    return evaluation_dict | {"schema": schema_dict}


def _describe(
    df: IntoFrame,
) -> dict[str, Any]:
    """Retrieve dictionary discribing dataframe.

    Mimics polars 'describe' method, but returns dict.
    """
    nw_df: NwDataFrame = nw.from_native(df)
    _validate_df_has_columns(nw_df)
    columns_to_check = nw_df.columns

    # Determine which columns should get std/mean/percentile statistics
    stat_cols = {c for c, dt in nw_df.schema.items() if dt.is_numeric()}

    required_exprs: list = [
        nw.lit("_^&^_".join(nw_df.columns)).alias("columns"),
    ]
    post_exprs: list = []
    required_exprs += [
        (nw.col(c).mean() if c in stat_cols else nw.lit(None)).alias(f"mean_{c}")
        for c in columns_to_check
    ]
    required_exprs += [
        (nw.col(c).std() if c in stat_cols else nw.lit(None)).alias(f"std_{c}")
        for c in columns_to_check
    ]
    required_exprs += [
        (nw.col(c).min() if c in stat_cols else nw.lit(None)).alias(f"min_{c}")
        for c in columns_to_check
    ]
    required_exprs += [
        (nw.col(c).max() if c in stat_cols else nw.lit(None)).alias(f"max_{c}")
        for c in columns_to_check
    ]
    required_exprs += [nw.lit(str(nw_df.schema[c])).alias(f"type_{c}") for c in columns_to_check]
    required_exprs += [nw.col(*columns_to_check).count().name.prefix("count_")]
    required_exprs += [
        nw.col(*columns_to_check).null_count().name.prefix("null_count_"),
    ]
    post_exprs += [
        (nw.col(f"null_count_{c}") / (nw.col(f"count_{c}") + nw.col(f"null_count_{c}"))).alias(
            f"null_percentage_{c}",
        )
        for c in columns_to_check
    ]
    post_exprs += [
        (
            nw.col(f"count_{columns_to_check[0]}") + nw.col(f"null_count_{columns_to_check[0]}")
        ).alias("length"),
    ]
    df_metrics = nw_df.select(
        *required_exprs,
    ).with_columns(*post_exprs)
    return _narwhals_to_dict(df_metrics)


def _profile_from_sampling(
    df: IntoFrame,
    samples: int = 100,
    n: int | None = None,
    fraction: int | None = None,
) -> list[dict[str, float]]:
    """Return profiles from sampling a single dataframe."""
    nw_df = nw.from_native(df)
    return [
        _describe(nw_df.sample(n=n, fraction=fraction, with_replacement=True))  # type: ignore[union-attr]
        for _ in range(samples)
    ]  # type: ignore[union-attr]


def _profile_from_samples(
    samples: Iterable[IntoFrame],
) -> list[dict[str, Any]]:
    """Return profiles from individual samples."""
    return [_describe(i) for i in samples]



from typing import Any, Literal

import pandas as pd
import polars as pl
import pytest

import wimsey


def dataframe(backend: Literal["pandas", "polars"]) -> Any:
    """Generate example dataframe for given backend"""
    data = {
        "first_name": ["Peter", "Jane", "Father", "Hercule", "Beatrice"],
        "last_name": ["Wimsey", "Marple", "Brown", "Poirot", "Bradley"],
        "rating": [9, 9, 7, 10, 8],
        "cases_solved": [11, 12, 53, 33, 66],
    }
    if backend == "pandas":
        return pd.DataFrame(data)
    if backend == "polars":
        return pl.DataFrame(data)
    msg = "Backend not recognised"
    raise NotImplementedError(msg)


@pytest.mark.parametrize(
    ("backend", "should_pass"),
    [
        ("pandas", True),
        ("pandas", False),
        ("polars", True),
        ("polars", False),
    ],
)
def test_validation_of_passing_example(backend: str, should_pass: bool) -> None:
    """Check that given dataframe passes example suite with validate function."""
    df = dataframe(backend)
    suite = f"tests/example-{'passing' if should_pass else 'failing'}.yaml"
    try:
        wimsey.validate(df, suite)
        assert should_pass, "Test passed"
    except wimsey.DataValidationError as error:
        assert not should_pass, f"Test failed: {error}"


@pytest.mark.parametrize(
    ("backend", "should_pass"),
    [
        ("pandas", True),
        ("pandas", False),
        ("polars", True),
        ("polars", False),
    ],
)
def test_test_of_passing_example(backend: str, should_pass: bool) -> None:
    """Check that given dataframe passes example suite with validate function."""
    df = dataframe(backend)
    suite = f"tests/example-{'passing' if should_pass else 'failing'}.yaml"
    results = wimsey.test(df, suite)
    assert results.success is should_pass, str(results.results)
    for result in results.results:
        assert result.success is should_pass, result.name



"""Functions relating to execution of defined tests."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import narwhals.stable.v1 as nw

from wimsey.config import _collect_tests, _read_config
from wimsey.dataframe import _evaluate
from wimsey.types import DataValidationError, FinalResult

if TYPE_CHECKING:
    from narwhals.stable.v1.typing import FrameT

    from wimsey.tests import Result
    from wimsey.types import GeneratedTest, _MagicExpr


def _run_all_tests(df: FrameT, tests: list[GeneratedTest]) -> FinalResult:
    """Run all given tests on a dataframe. Will return a `FinalResult` object."""
    metrics: list[nw.Expr | _MagicExpr] = []
    for i, expr_tuple in enumerate(tests):
        expr, _ = expr_tuple
        if isinstance(expr, nw.Expr):
            metrics.append(expr.alias(str(i)))
    evaluation: dict[str, Any] = _evaluate(
        df,
        metrics=metrics,
    )
    results: list[Result] = []
    for i, expr_tuple in enumerate(tests):
        expr, check = expr_tuple
        results.append(
            check(
                evaluation[str(i)] if isinstance(expr, nw.Expr) else evaluation[expr.expr_name],
            ),
        )
    return FinalResult(
        success=all(i.success for i in results),
        results=results,
    )


def test(
    df: FrameT,
    contract: str | list[dict] | dict,
    storage_options: dict | None = None,
) -> FinalResult:
    """Test a dataframe against a data contract.

    Carry out tests on dataframe and return results. This will *not* raise
    an exception on test failure, and will instead return a 'final_result'
    object, with a boolean 'success' field, and a detailed list of individual
    tests.

    If you want to halt processing in the event of a data contract failure,
    see `validate` function.
    """
    tests = (
        _read_config(path=contract, storage_options=storage_options)
        if isinstance(contract, str)
        else _collect_tests(contract)
    )
    return _run_all_tests(df, tests)


def validate(
    df: FrameT,
    contract: str | list[dict] | dict,
    storage_options: dict | None = None,
) -> FrameT:
    """Validate a dataframe against a data contract.

    Carry out tests on dataframe, returning original dataframe if tests are
    successful, and raising a DataValidationException in case of failure.
    """
    results = test(
        df=df,
        contract=contract,
        storage_options=storage_options,
    )
    if not results.success:
        failures: list[str] = [
            f"{i.name} (unexpected: {i.unexpected})" for i in results.results if not i.success
        ]
        newline = "\n - "
        msg = f"At least one test failed:\n - {newline.join(failures)}"
        raise DataValidationError(msg)
    return df
